@@ -1,6 +1,5 @@
 use crate::store::{StoreData, StoreOpaque, Stored};
 use crate::trampoline::{generate_global_export, generate_table_export};
-use crate::values::{from_checked_anyfunc, into_checked_anyfunc};
 use crate::{
     AsContext, AsContextMut, ExternRef, ExternType, Func, GlobalType, Instance, Memory, Module,
     Mutability, TableType, Trap, Val, ValType,
@@ -307,7 +306,7 @@ impl Global {
                         .map(|inner| ExternRef { inner }),
                 ),
                 ValType::FuncRef => {
-                    from_checked_anyfunc(definition.as_anyfunc() as *mut _, store.0)
+                    Val::FuncRef(Func::from_raw(store, definition.as_anyfunc() as usize))
                 }
                 ValType::V128 => Val::V128(*definition.as_u128()),
             }
@@ -407,6 +406,14 @@ impl Table {
     /// Returns an error if `init` does not match the element type of the table,
     /// or if `init` does not belong to the `store` provided.
     ///
+    /// # Panics
+    ///
+    /// This function will panic when used with a [`Store`](`crate::Store`)
+    /// which has a [`ResourceLimiterAsync`](`crate::ResourceLimiterAsync`)
+    /// (see also: [`Store::limiter_async`](`crate::Store::limiter_async`).
+    /// When using an async resource limiter, use [`Table::new_async`]
+    /// instead.
+    ///
     /// # Examples
     ///
     /// ```
@@ -437,29 +444,37 @@ impl Table {
         Table::_new(store.as_context_mut().0, ty, init)
     }
 
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
+    /// Async variant of [`Table::new`]. You must use this variant with
+    /// [`Store`](`crate::Store`)s which have a
+    /// [`ResourceLimiterAsync`](`crate::ResourceLimiterAsync`).
+    ///
+    /// # Panics
+    ///
+    /// This function will panic when used with a non-async
+    /// [`Store`](`crate::Store`)
+    #[cfg(feature = "async")]
+    pub async fn new_async<T>(
+        mut store: impl AsContextMut<Data = T>,
+        ty: TableType,
+        init: Val,
+    ) -> Result<Table>
+    where
+        T: Send,
+    {
+        let mut store = store.as_context_mut();
+        assert!(
+            store.0.async_support(),
+            "cannot use `new_async` without enabling async support on the config"
+        );
+        store
+            .on_fiber(|store| Table::_new(store.0, ty, init))
+            .await?
+    }
+
     fn _new(store: &mut StoreOpaque, ty: TableType, init: Val) -> Result<Table> {
-        if init.ty() != ty.element() {
-            bail!(
-                "table initialization value type {:?} does not have expected type {:?}",
-                init.ty(),
-                ty.element(),
-            );
-        }
         let wasmtime_export = generate_table_export(store, &ty)?;
-
-        let init: runtime::TableElement = match ty.element() {
-            ValType::FuncRef => into_checked_anyfunc(init, store)?.into(),
-            ValType::ExternRef => init
-                .externref()
-                .ok_or_else(|| {
-                    anyhow!("table initialization value does not have expected type `externref`")
-                })?
-                .map(|x| x.inner)
-                .into(),
-            ty => bail!("unsupported table element type: {:?}", ty),
-        };
-
-        // Initialize entries with the init value.
+        let init = init.into_table_element(store, ty.element())?;
         unsafe {
             let table = Table::from_wasmtime_table(wasmtime_export, store);
             (*table.wasmtime_table(store))
@@ -503,7 +518,10 @@ impl Table {
         let table = self.wasmtime_table(store);
         unsafe {
             match (*table).get(index)? {
-                runtime::TableElement::FuncRef(f) => Some(from_checked_anyfunc(f, store)),
+                runtime::TableElement::FuncRef(f) => {
+                    let func = Func::from_caller_checked_anyfunc(store, f);
+                    Some(Val::FuncRef(func))
+                }
                 runtime::TableElement::ExternRef(None) => Some(Val::ExternRef(None)),
                 runtime::TableElement::ExternRef(Some(x)) => {
                     Some(Val::ExternRef(Some(ExternRef { inner: x })))
@@ -541,7 +559,10 @@ impl Table {
     ///
     /// Panics if `store` does not own this table.
     pub fn size(&self, store: impl AsContext) -> u32 {
-        let store = store.as_context();
+        self.internal_size(store.as_context().0)
+    }
+
+    pub(crate) fn internal_size(&self, store: &StoreOpaque) -> u32 {
         unsafe { (*store[self.0].definition).current_elements }
     }
 
@@ -560,13 +581,19 @@ impl Table {
     /// # Panics
     ///
     /// Panics if `store` does not own this table.
+    ///
+    /// This function will panic when used with a [`Store`](`crate::Store`)
+    /// which has a [`ResourceLimiterAsync`](`crate::ResourceLimiterAsync`)
+    /// (see also: [`Store::limiter_async`](`crate::Store::limiter_async`)).
+    /// When using an async resource limiter, use [`Table::grow_async`]
+    /// instead.
     pub fn grow(&self, mut store: impl AsContextMut, delta: u32, init: Val) -> Result<u32> {
         let store = store.as_context_mut().0;
         let ty = self.ty(&store).element().clone();
         let init = init.into_table_element(store, ty)?;
         let table = self.wasmtime_table(store);
         unsafe {
-            match (*table).grow(delta, init, store.limiter()) {
+            match (*table).grow(delta, init, store)? {
                 Some(size) => {
                     let vm = (*table).vmtable();
                     *store[self.0].definition = vm;
@@ -577,6 +604,34 @@ impl Table {
         }
     }
 
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
+    /// Async variant of [`Table::grow`]. Required when using a
+    /// [`ResourceLimiterAsync`](`crate::ResourceLimiterAsync`).
+    ///
+    /// # Panics
+    ///
+    /// This function will panic when used with a non-async
+    /// [`Store`](`crate::Store`).
+    #[cfg(feature = "async")]
+    pub async fn grow_async<T>(
+        &self,
+        mut store: impl AsContextMut<Data = T>,
+        delta: u32,
+        init: Val,
+    ) -> Result<u32>
+    where
+        T: Send,
+    {
+        let mut store = store.as_context_mut();
+        assert!(
+            store.0.async_support(),
+            "cannot use `grow_async` without enabling async support on the config"
+        );
+        store
+            .on_fiber(|store| self.grow(store, delta, init))
+            .await?
+    }
+
     /// Copy `len` elements from `src_table[src_index..]` into
     /// `dst_table[dst_index..]`.
     ///
