@@ -1,12 +1,9 @@
 use crate::store::{StoreData, StoreOpaque, Stored};
 use crate::{
     AsContext, AsContextMut, CallHook, Engine, Extern, FuncType, Instance, InterruptHandle,
-    StoreContext, StoreContextMut, Trap, Val, ValType,
+    StoreContext, StoreContextMut, Trap, Val, ValRaw, ValType,
 };
 use anyhow::{bail, Context as _, Result};
-use std::cmp::max;
-use std::error::Error;
-use std::fmt;
 use std::future::Future;
 use std::mem;
 use std::panic::{self, AssertUnwindSafe};
@@ -90,13 +87,13 @@ use wasmtime_runtime::{
 ///
 /// // Work with `foo` as a `Func` at this point, such as calling it
 /// // dynamically...
-/// match foo.call(&mut store, &[]) {
-///     Ok(result) => { /* ... */ }
+/// match foo.call(&mut store, &[], &mut []) {
+///     Ok(()) => { /* ... */ }
 ///     Err(trap) => {
 ///         panic!("execution of `foo` resulted in a wasm trap: {}", trap);
 ///     }
 /// }
-/// foo.call(&mut store, &[])?;
+/// foo.call(&mut store, &[], &mut [])?;
 ///
 /// // ... or we can make a static assertion about its signature and call it.
 /// // Our first call here can fail if the signatures don't match, and then the
@@ -184,9 +181,14 @@ use wasmtime_runtime::{
 #[repr(transparent)] // here for the C API
 pub struct Func(Stored<FuncData>);
 
+pub(crate) struct FuncData {
+    kind: FuncKind,
+    ty: FuncType,
+}
+
 /// The three ways that a function can be created and referenced from within a
 /// store.
-pub(crate) enum FuncData {
+enum FuncKind {
     /// A function already owned by the store via some other means. This is
     /// used, for example, when creating a `Func` from an instance's exported
     /// function. The instance's `InstanceHandle` is already owned by the store
@@ -302,20 +304,48 @@ impl Func {
     #[cfg(compiler)]
     #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn new<T>(
-        mut store: impl AsContextMut<Data = T>,
+        store: impl AsContextMut<Data = T>,
         ty: FuncType,
         func: impl Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<(), Trap> + Send + Sync + 'static,
     ) -> Self {
-        let store = store.as_context_mut().0;
-
-        // part of this unsafety is about matching the `T` to a `Store<T>`,
-        // which is done through the `AsContextMut` bound above.
+        let ty_clone = ty.clone();
         unsafe {
-            let host = HostFunc::new(store.engine(), ty, func);
-            host.into_func(store)
+            Func::new_unchecked(store, ty, move |caller, values| {
+                Func::invoke(caller, &ty_clone, values, &func)
+            })
         }
     }
 
+    /// Creates a new [`Func`] with the given arguments, although has fewer
+    /// runtime checks than [`Func::new`].
+    ///
+    /// This function takes a callback of a different signature than
+    /// [`Func::new`], instead receiving a raw pointer with a list of [`ValRaw`]
+    /// structures. These values have no type information associated with them
+    /// so it's up to the caller to provide a function that will correctly
+    /// interpret the list of values as those coming from the `ty` specified.
+    ///
+    /// If you're calling this from Rust it's recommended to either instead use
+    /// [`Func::new`] or [`Func::wrap`]. The [`Func::wrap`] API, in particular,
+    /// is both safer and faster than this API.
+    ///
+    /// # Unsafety
+    ///
+    /// This function is not safe because it's not known at compile time that
+    /// the `func` provided correctly interprets the argument types provided to
+    /// it, or that the results it produces will be of the correct type.
+    #[cfg(compiler)]
+    #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
+    pub unsafe fn new_unchecked<T>(
+        mut store: impl AsContextMut<Data = T>,
+        ty: FuncType,
+        func: impl Fn(Caller<'_, T>, *mut ValRaw) -> Result<(), Trap> + Send + Sync + 'static,
+    ) -> Self {
+        let store = store.as_context_mut().0;
+        let host = HostFunc::new_unchecked(store.engine(), ty, func);
+        host.into_func(store)
+    }
+
     /// Creates a new host-defined WebAssembly function which, when called,
     /// will run the asynchronous computation defined by `func` to completion
     /// and then return the result to WebAssembly.
@@ -408,9 +438,9 @@ impl Func {
 
     pub(crate) unsafe fn from_caller_checked_anyfunc(
         store: &mut StoreOpaque,
-        anyfunc: *mut VMCallerCheckedAnyfunc,
-    ) -> Option<Self> {
-        let anyfunc = NonNull::new(anyfunc)?;
+        raw: *mut VMCallerCheckedAnyfunc,
+    ) -> Option<Func> {
+        let anyfunc = NonNull::new(raw)?;
         debug_assert!(anyfunc.as_ref().type_index != VMSharedSignatureIndex::default());
         let export = ExportFunction { anyfunc };
         Some(Func::from_wasmtime_function(export, store))
@@ -647,42 +677,115 @@ impl Func {
     ///
     /// Panics if `store` does not own this function.
     pub fn ty(&self, store: impl AsContext) -> FuncType {
-        // Signatures should always be registered in the engine's registry of
-        // shared signatures, so we should be able to unwrap safely here.
-        let store = store.as_context();
-        let sig_index = unsafe { store[self.0].export().anyfunc.as_ref().type_index };
-        FuncType::from_wasm_func_type(
-            store
-                .engine()
-                .signatures()
-                .lookup_type(sig_index)
-                .expect("signature should be registered"),
-        )
+        store.as_context()[self.0].ty.clone()
     }
 
     pub(crate) fn sig_index(&self, data: &StoreData) -> VMSharedSignatureIndex {
         unsafe { data[self.0].export().anyfunc.as_ref().type_index }
     }
 
-    /// Invokes this function with the `params` given, returning the results and
-    /// any trap, if one occurs.
+    /// Invokes this function with the `params` given and writes returned values
+    /// to `results`.
     ///
     /// The `params` here must match the type signature of this `Func`, or a
     /// trap will occur. If a trap occurs while executing this function, then a
-    /// trap will also be returned.
+    /// trap will also be returned. Additionally `results` must have the same
+    /// length as the number of results for this function.
     ///
     /// # Panics
     ///
     /// This function will panic if called on a function belonging to an async
     /// store. Asynchronous stores must always use `call_async`.
     /// initiates a panic. Also panics if `store` does not own this function.
-    pub fn call(&self, mut store: impl AsContextMut, params: &[Val]) -> Result<Box<[Val]>> {
+    pub fn call(
+        &self,
+        mut store: impl AsContextMut,
+        params: &[Val],
+        results: &mut [Val],
+    ) -> Result<()> {
         assert!(
             !store.as_context().async_support(),
             "must use `call_async` when async support is enabled on the config",
         );
-        let my_ty = self.ty(&store);
-        self.call_impl(&mut store.as_context_mut(), my_ty, params)
+        self.call_impl(&mut store.as_context_mut(), params, results)
+    }
+
+    /// Invokes this function in an "unchecked" fashion, reading parameters and
+    /// writing results to `params_and_returns`.
+    ///
+    /// This function is the same as [`Func::call`] except that the arguments
+    /// and results both use a different representation. If possible it's
+    /// recommended to use [`Func::call`] if safety isn't necessary or to use
+    /// [`Func::typed`] in conjunction with [`TypedFunc::call`] since that's
+    /// both safer and faster than this method of invoking a function.
+    ///
+    /// Note that if this function takes `externref` arguments then it will
+    /// **not** automatically GC unlike the [`Func::call`] and
+    /// [`TypedFunc::call`] functions. This means that if this function is
+    /// invoked many times with new `ExternRef` values and no other GC happens
+    /// via any other means then no values will get collected.
+    ///
+    /// # Unsafety
+    ///
+    /// This function is unsafe because the `params_and_returns` argument is not
+    /// validated at all. It must uphold invariants such as:
+    ///
+    /// * It's a valid pointer to an array
+    /// * It has enough space to store all parameters
+    /// * It has enough space to store all results (not at the same time as
+    ///   parameters)
+    /// * Parameters are initially written to the array and have the correct
+    ///   types and such.
+    /// * Reference types like `externref` and `funcref` are valid at the
+    ///   time of this call and for the `store` specified.
+    ///
+    /// These invariants are all upheld for you with [`Func::call`] and
+    /// [`TypedFunc::call`].
+    pub unsafe fn call_unchecked(
+        &self,
+        mut store: impl AsContextMut,
+        params_and_returns: *mut ValRaw,
+    ) -> Result<(), Trap> {
+        let mut store = store.as_context_mut();
+        let data = &store.0.store_data()[self.0];
+        let trampoline = data.trampoline();
+        let anyfunc = data.export().anyfunc;
+        invoke_wasm_and_catch_traps(&mut store, |callee| {
+            trampoline(
+                (*anyfunc.as_ptr()).vmctx,
+                callee,
+                (*anyfunc.as_ptr()).func_ptr.as_ptr(),
+                params_and_returns,
+            )
+        })
+    }
+
+    /// Converts the raw representation of a `funcref` into an `Option<Func>`
+    ///
+    /// This is intended to be used in conjunction with [`Func::new_unchecked`],
+    /// [`Func::call_unchecked`], and [`ValRaw`] with its `funcref` field.
+    ///
+    /// # Unsafety
+    ///
+    /// This function is not safe because `raw` is not validated at all. The
+    /// caller must guarantee that `raw` is owned by the `store` provided and is
+    /// valid within the `store`.
+    pub unsafe fn from_raw(mut store: impl AsContextMut, raw: usize) -> Option<Func> {
+        Func::from_caller_checked_anyfunc(store.as_context_mut().0, raw as *mut _)
+    }
+
+    /// Extracts the raw value of this `Func`, which is owned by `store`.
+    ///
+    /// This function returns a value that's suitable for writing into the
+    /// `funcref` field of the [`ValRaw`] structure.
+    ///
+    /// # Unsafety
+    ///
+    /// The returned value is only valid for as long as the store is alive and
+    /// this function is properly rooted within it. Additionally this function
+    /// should not be liberally used since it's a very low-level knob.
+    pub unsafe fn to_raw(&self, store: impl AsContext) -> usize {
+        self.caller_checked_anyfunc(store.as_context().0).as_ptr() as usize
     }
 
     /// Invokes this function with the `params` given, returning the results
@@ -713,7 +816,8 @@ impl Func {
         &self,
         mut store: impl AsContextMut<Data = T>,
         params: &[Val],
-    ) -> Result<Box<[Val]>>
+        results: &mut [Val],
+    ) -> Result<()>
     where
         T: Send,
     {
@@ -722,9 +826,8 @@ impl Func {
             store.0.async_support(),
             "cannot use `call_async` without enabling async support in the config",
         );
-        let my_ty = self.ty(&store);
         let result = store
-            .on_fiber(|store| self.call_impl(store, my_ty, params))
+            .on_fiber(|store| self.call_impl(store, params, results))
             .await??;
         Ok(result)
     }
@@ -732,94 +835,81 @@ impl Func {
     fn call_impl<T>(
         &self,
         store: &mut StoreContextMut<'_, T>,
-        my_ty: FuncType,
         params: &[Val],
-    ) -> Result<Box<[Val]>> {
-        let mut values_vec = write_params(store.0, &my_ty, params)?;
-
-        // Call the trampoline.
-        unsafe {
-            let data = &store.0.store_data()[self.0];
-            let trampoline = data.trampoline();
-            let anyfunc = data.export().anyfunc;
-            invoke_wasm_and_catch_traps(store, |callee| {
-                trampoline(
-                    (*anyfunc.as_ptr()).vmctx,
-                    callee,
-                    (*anyfunc.as_ptr()).func_ptr.as_ptr(),
-                    values_vec.as_mut_ptr(),
-                )
-            })?;
+        results: &mut [Val],
+    ) -> Result<()> {
+        // We need to perform a dynamic check that the arguments given to us
+        // match the signature of this function and are appropriate to pass to
+        // this function. This involves checking to make sure we have the right
+        // number and types of arguments as well as making sure everything is
+        // from the same `Store`.
+        let ty = &store[self.0].ty;
+        if ty.params().len() != params.len() {
+            bail!(
+                "expected {} arguments, got {}",
+                ty.params().len(),
+                params.len()
+            );
         }
-
-        return Ok(read_results(store.0, &my_ty, &values_vec));
-
-        fn write_params(
-            store: &mut StoreOpaque,
-            ty: &FuncType,
-            params: &[Val],
-        ) -> Result<Vec<u128>> {
-            // We need to perform a dynamic check that the arguments given to us
-            // match the signature of this function and are appropriate to pass to
-            // this function. This involves checking to make sure we have the right
-            // number and types of arguments as well as making sure everything is
-            // from the same `Store`.
-            if ty.params().len() != params.len() {
+        if ty.results().len() != results.len() {
+            bail!(
+                "expected {} results, got {}",
+                ty.results().len(),
+                results.len()
+            );
+        }
+        for (ty, arg) in ty.params().zip(params) {
+            if arg.ty() != ty {
                 bail!(
-                    "expected {} arguments, got {}",
-                    ty.params().len(),
-                    params.len()
+                    "argument type mismatch: found {} but expected {}",
+                    arg.ty(),
+                    ty
                 );
             }
-
-            let mut values_vec = vec![0; max(params.len(), ty.results().len())];
-
-            // Whenever we pass `externref`s from host code to Wasm code, they
-            // go into the `VMExternRefActivationsTable`. But the table might be
-            // at capacity already, so check for that. If it is at capacity
-            // (unlikely) then do a GC to free up space. This is necessary
-            // because otherwise we would either keep filling up the bump chunk
-            // and making it larger and larger or we would always take the slow
-            // path when inserting references into the table.
-            if ty.as_wasm_func_type().externref_params_count()
-                > store
-                    .externref_activations_table()
-                    .bump_capacity_remaining()
-            {
-                store.gc();
+            if !arg.comes_from_same_store(store.0) {
+                bail!("cross-`Store` values are not currently supported");
             }
-
-            // Store the argument values into `values_vec`.
-            let param_tys = ty.params();
-            for ((arg, slot), ty) in params.iter().cloned().zip(&mut values_vec).zip(param_tys) {
-                if arg.ty() != ty {
-                    bail!(
-                        "argument type mismatch: found {} but expected {}",
-                        arg.ty(),
-                        ty
-                    );
-                }
-                if !arg.comes_from_same_store(store) {
-                    bail!("cross-`Store` values are not currently supported");
-                }
-                unsafe {
-                    arg.write_value_without_gc(store, slot);
-                }
-            }
-
-            Ok(values_vec)
         }
 
-        fn read_results(store: &mut StoreOpaque, ty: &FuncType, values_vec: &[u128]) -> Box<[Val]> {
-            let mut results = Vec::with_capacity(ty.results().len());
-            for (index, ty) in ty.results().enumerate() {
-                unsafe {
-                    let ptr = &values_vec[index];
-                    results.push(Val::read_value_from(store, ptr, ty));
-                }
-            }
-            results.into()
+        let values_vec_size = params.len().max(ty.results().len());
+
+        // Whenever we pass `externref`s from host code to Wasm code, they
+        // go into the `VMExternRefActivationsTable`. But the table might be
+        // at capacity already, so check for that. If it is at capacity
+        // (unlikely) then do a GC to free up space. This is necessary
+        // because otherwise we would either keep filling up the bump chunk
+        // and making it larger and larger or we would always take the slow
+        // path when inserting references into the table.
+        if ty.as_wasm_func_type().externref_params_count()
+            > store
+                .0
+                .externref_activations_table()
+                .bump_capacity_remaining()
+        {
+            store.gc();
         }
+
+        // Store the argument values into `values_vec`.
+        let mut values_vec = store.0.take_wasm_val_raw_storage();
+        debug_assert!(values_vec.is_empty());
+        values_vec.resize_with(values_vec_size, || ValRaw { i32: 0 });
+        for (arg, slot) in params.iter().cloned().zip(&mut values_vec) {
+            unsafe {
+                *slot = arg.to_raw(&mut *store);
+            }
+        }
+
+        unsafe {
+            self.call_unchecked(&mut *store, values_vec.as_mut_ptr())?;
+        }
+
+        for ((i, slot), val) in results.iter_mut().enumerate().zip(&values_vec) {
+            let ty = store[self.0].ty.results().nth(i).unwrap();
+            *slot = unsafe { Val::from_raw(&mut *store, *val, ty) };
+        }
+        values_vec.truncate(0);
+        store.0.save_wasm_val_raw_storage(values_vec);
+        Ok(())
     }
 
     #[inline]
@@ -836,8 +926,21 @@ impl Func {
     ) -> Self {
         let anyfunc = export.anyfunc.as_ref();
         let trampoline = store.lookup_trampoline(&*anyfunc);
-        let data = FuncData::StoreOwned { trampoline, export };
-        Func(store.store_data_mut().insert(data))
+        Func::from_func_kind(FuncKind::StoreOwned { trampoline, export }, store)
+    }
+
+    fn from_func_kind(kind: FuncKind, store: &mut StoreOpaque) -> Self {
+        // Signatures should always be registered in the engine's registry of
+        // shared signatures, so we should be able to unwrap safely here.
+        let ty = unsafe { kind.export().anyfunc.as_ref().type_index };
+        let ty = FuncType::from_wasm_func_type(
+            store
+                .engine()
+                .signatures()
+                .lookup_type(ty)
+                .expect("signature should be registered"),
+        );
+        Func(store.store_data_mut().insert(FuncData { kind, ty }))
     }
 
     pub(crate) fn vmimport(&self, store: &mut StoreOpaque) -> VMFunctionImport {
@@ -857,7 +960,7 @@ impl Func {
     fn invoke<T>(
         mut caller: Caller<'_, T>,
         ty: &FuncType,
-        values_vec: *mut u128,
+        values_vec: *mut ValRaw,
         func: &dyn Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<(), Trap>,
     ) -> Result<(), Trap> {
         caller.store.0.call_hook(CallHook::CallingHost)?;
@@ -876,10 +979,7 @@ impl Func {
         let nparams = ty.params().len();
         val_vec.reserve(nparams + ty.results().len());
         for (i, ty) in ty.params().enumerate() {
-            unsafe {
-                let val = Val::read_value_from(caller.store.0, values_vec.add(i), ty);
-                val_vec.push(val);
-            }
+            val_vec.push(unsafe { Val::from_raw(&mut caller.store, *values_vec.add(i), ty) })
         }
 
         val_vec.extend((0..ty.results().len()).map(|_| Val::null()));
@@ -913,7 +1013,7 @@ impl Func {
                 ));
             }
             unsafe {
-                ret.write_value_without_gc(caller.store.0, values_vec.add(i));
+                *values_vec.add(i) = ret.to_raw(&mut caller.store);
             }
         }
 
@@ -1027,9 +1127,7 @@ impl Func {
     /// and similarly if a function has multiple results you can bind that too
     ///
     /// ```
-    /// # #[cfg(not(feature = "old-x86-backend"))]
     /// # use wasmtime::*;
-    /// # #[cfg(not(feature = "old-x86-backend"))]
     /// # fn foo(add_with_overflow: &Func, mut store: Store<()>) -> anyhow::Result<()> {
     /// let typed = add_with_overflow.typed::<(u32, u32), (u32, i32), _>(&store)?;
     /// let (result, overflow) = typed.call(&mut store, (u32::max_value(), 2))?;
@@ -1243,7 +1341,7 @@ pub unsafe trait WasmRet {
     fn func_type(params: impl Iterator<Item = ValType>) -> FuncType;
 
     #[doc(hidden)]
-    unsafe fn wrap_trampoline(ptr: *mut u128, f: impl FnOnce(Self::Retptr) -> Self::Abi);
+    unsafe fn wrap_trampoline(ptr: *mut ValRaw, f: impl FnOnce(Self::Retptr) -> Self::Abi);
 
     // Utilities used to convert an instance of this type to a `Result`
     // explicitly, used when wrapping async functions which always bottom-out
@@ -1280,7 +1378,7 @@ where
         FuncType::new(params, Some(<Self as WasmTy>::valtype()))
     }
 
-    unsafe fn wrap_trampoline(ptr: *mut u128, f: impl FnOnce(Self::Retptr) -> Self::Abi) {
+    unsafe fn wrap_trampoline(ptr: *mut ValRaw, f: impl FnOnce(Self::Retptr) -> Self::Abi) {
         *ptr.cast::<Self::Abi>() = f(());
     }
 
@@ -1320,7 +1418,7 @@ where
         T::func_type(params)
     }
 
-    unsafe fn wrap_trampoline(ptr: *mut u128, f: impl FnOnce(Self::Retptr) -> Self::Abi) {
+    unsafe fn wrap_trampoline(ptr: *mut ValRaw, f: impl FnOnce(Self::Retptr) -> Self::Abi) {
         T::wrap_trampoline(ptr, f)
     }
 
@@ -1366,7 +1464,7 @@ macro_rules! impl_wasm_host_results {
             }
 
             #[allow(unused_assignments)]
-            unsafe fn wrap_trampoline(mut _ptr: *mut u128, f: impl FnOnce(Self::Retptr) -> Self::Abi) {
+            unsafe fn wrap_trampoline(mut _ptr: *mut ValRaw, f: impl FnOnce(Self::Retptr) -> Self::Abi) {
                 let ($($t,)*) = <($($t::Abi,)*) as HostAbi>::call(f);
                 $(
                     *_ptr.cast() = $t;
@@ -1462,12 +1560,10 @@ macro_rules! impl_host_abi {
         #[doc(hidden)]
         #[allow(non_snake_case)]
         #[repr(C)]
-        #[cfg(not(feature = "old-x86-backend"))]
         pub struct [<TupleRet $n>]<$($u,)*> {
             $($u: $u,)*
         }
 
-        #[cfg(not(feature = "old-x86-backend"))]
         #[allow(non_snake_case, unused_assignments)]
         impl<$t: Copy, $($u: Copy,)*> HostAbi for ($t, $($u,)*) {
             type Abi = $t;
@@ -1686,25 +1782,6 @@ impl<T> AsContextMut for Caller<'_, T> {
     }
 }
 
-fn cross_store_trap() -> Box<dyn Error + Send + Sync> {
-    #[derive(Debug)]
-    struct CrossStoreError;
-
-    impl Error for CrossStoreError {}
-
-    impl fmt::Display for CrossStoreError {
-        fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-            write!(
-                f,
-                "host function attempted to return cross-`Store` \
-                 value to Wasm",
-            )
-        }
-    }
-
-    Box::new(CrossStoreError)
-}
-
 macro_rules! impl_into_func {
     ($num:tt $($args:ident)*) => {
         // Implement for functions without a leading `&Caller` parameter,
@@ -1754,7 +1831,7 @@ macro_rules! impl_into_func {
                 {
                     enum CallResult<U> {
                         Ok(U),
-                        Trap(Box<dyn Error + Send + Sync>),
+                        Trap(anyhow::Error),
                         Panic(Box<dyn std::any::Any + Send>),
                     }
 
@@ -1803,7 +1880,7 @@ macro_rules! impl_into_func {
                                 // can't assume it returned a value that is
                                 // compatible with this store.
                                 if !ret.compatible_with_store(caller.store.0) {
-                                    CallResult::Trap(cross_store_trap())
+                                    CallResult::Trap(anyhow::anyhow!("host function attempted to return cross-`Store` value to Wasm"))
                                 } else {
                                     match ret.into_abi_for_ret(caller.store.0, retptr) {
                                         Ok(val) => CallResult::Ok(val),
@@ -1833,7 +1910,7 @@ macro_rules! impl_into_func {
                     callee_vmctx: *mut VMContext,
                     caller_vmctx: *mut VMContext,
                     ptr: *const VMFunctionBody,
-                    args: *mut u128,
+                    args: *mut ValRaw,
                 )
                 where
                     $($args: WasmTy,)*
@@ -1923,14 +2000,23 @@ impl HostFunc {
         func: impl Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<(), Trap> + Send + Sync + 'static,
     ) -> Self {
         let ty_clone = ty.clone();
-
-        // Create a trampoline that converts raw u128 values to `Val`
-        let func = move |caller_vmctx, values_vec: *mut u128| unsafe {
-            Caller::with(caller_vmctx, |caller| {
-                Func::invoke(caller, &ty_clone, values_vec, &func)
+        unsafe {
+            HostFunc::new_unchecked(engine, ty, move |caller, values| {
+                Func::invoke(caller, &ty_clone, values, &func)
             })
-        };
+        }
+    }
 
+    /// Analog of [`Func::new_unchecked`]
+    #[cfg(compiler)]
+    pub unsafe fn new_unchecked<T>(
+        engine: &Engine,
+        ty: FuncType,
+        func: impl Fn(Caller<'_, T>, *mut ValRaw) -> Result<(), Trap> + Send + Sync + 'static,
+    ) -> Self {
+        let func = move |caller_vmctx, values: *mut ValRaw| unsafe {
+            Caller::<T>::with(caller_vmctx, |caller| func(caller, values))
+        };
         let (instance, trampoline) = crate::trampoline::create_function(&ty, func, engine)
             .expect("failed to create function");
         HostFunc::_new(engine, instance, trampoline)
@@ -1972,13 +2058,13 @@ impl HostFunc {
     pub unsafe fn to_func(self: &Arc<Self>, store: &mut StoreOpaque) -> Func {
         self.register_trampoline(store);
         let me = self.clone();
-        Func(store.store_data_mut().insert(FuncData::SharedHost(me)))
+        Func::from_func_kind(FuncKind::SharedHost(me), store)
     }
 
     /// Same as [`HostFunc::to_func`], different ownership.
     unsafe fn into_func(self, store: &mut StoreOpaque) -> Func {
         self.register_trampoline(store);
-        Func(store.store_data_mut().insert(FuncData::Host(self)))
+        Func::from_func_kind(FuncKind::Host(self), store)
     }
 
     unsafe fn register_trampoline(&self, store: &mut StoreOpaque) {
@@ -2014,20 +2100,28 @@ impl Drop for HostFunc {
 }
 
 impl FuncData {
+    #[inline]
     fn trampoline(&self) -> VMTrampoline {
-        match self {
-            FuncData::StoreOwned { trampoline, .. } => *trampoline,
-            FuncData::SharedHost(host) => host.trampoline,
-            FuncData::Host(host) => host.trampoline,
+        match &self.kind {
+            FuncKind::StoreOwned { trampoline, .. } => *trampoline,
+            FuncKind::SharedHost(host) => host.trampoline,
+            FuncKind::Host(host) => host.trampoline,
         }
     }
 
+    #[inline]
+    fn export(&self) -> &ExportFunction {
+        self.kind.export()
+    }
+}
+
+impl FuncKind {
     #[inline]
     fn export(&self) -> &ExportFunction {
         match self {
-            FuncData::StoreOwned { export, .. } => export,
-            FuncData::SharedHost(host) => &host.export,
-            FuncData::Host(host) => &host.export,
+            FuncKind::StoreOwned { export, .. } => export,
+            FuncKind::SharedHost(host) => &host.export,
+            FuncKind::Host(host) => &host.export,
         }
     }
 }
