@@ -8,7 +8,7 @@ use std::fs;
 use std::mem;
 use std::path::Path;
 use std::sync::Arc;
-use wasmparser::Validator;
+use wasmparser::{Parser, ValidPayload, Validator};
 use wasmtime_environ::{ModuleEnvironment, ModuleIndex, PrimaryMap};
 use wasmtime_jit::{CompiledModule, CompiledModuleInfo, MmapVec, TypeTables};
 
@@ -398,13 +398,10 @@ impl Module {
                 .collect();
 
             let mut obj = engine.compiler().object()?;
-            let (funcs, trampolines) = engine.compiler().emit_obj(
-                &translation,
-                &types,
-                funcs,
-                tunables.generate_native_debuginfo,
-                &mut obj,
-            )?;
+            let (funcs, trampolines) =
+                engine
+                    .compiler()
+                    .emit_obj(&translation, &types, funcs, tunables, &mut obj)?;
 
             // If configured, attempt to use paged memory initialization
             // instead of the default mode of memory initialization
@@ -435,6 +432,10 @@ impl Module {
     /// [`Module::serialize`] and [`Engine::precompile_module`] back into an
     /// in-memory [`Module`] that's ready to be instantiated.
     ///
+    /// Note that the [`Module::deserialize_file`] method is more optimized than
+    /// this function, so if the serialized module is already present in a file
+    /// it's recommended to use that method instead.
+    ///
     /// # Unsafety
     ///
     /// This function is marked as `unsafe` because if fed invalid input or used
@@ -500,7 +501,9 @@ impl Module {
         let signatures = Arc::new(SignatureCollection::new_for_module(
             engine.signatures(),
             &types.wasm_signatures,
-            modules.iter().flat_map(|m| m.trampolines()),
+            modules
+                .iter()
+                .flat_map(|m| m.trampolines().map(|(idx, f, _)| (idx, f))),
         ));
 
         let module = modules.remove(main_module);
@@ -591,7 +594,15 @@ impl Module {
     pub fn validate(engine: &Engine, binary: &[u8]) -> Result<()> {
         let mut validator = Validator::new();
         validator.wasm_features(engine.config().features);
-        validator.validate_all(binary)?;
+
+        let mut functions = Vec::new();
+        for payload in Parser::new(0).parse_all(binary) {
+            if let ValidPayload::Func(a, b) = validator.payload(&payload?)? {
+                functions.push((a, b));
+            }
+        }
+
+        engine.run_maybe_parallel(functions, |(mut validator, body)| validator.validate(&body))?;
         Ok(())
     }
 
@@ -612,10 +623,15 @@ impl Module {
         sig
     }
 
-    /// Serialize the module to a vector of bytes.
+    /// Serializes this module to a vector of bytes.
     ///
-    /// Use `Module::new` or `Module::from_binary` to create the module
-    /// from the bytes.
+    /// This function is similar to the [`Engine::precompile_module`] method
+    /// where it produces an artifact of Wasmtime which is suitable to later
+    /// pass into [`Module::deserialize`]. If a module is never instantiated
+    /// then it's recommended to use [`Engine::precompile_module`] instead of
+    /// this method, but if a module is both instantiated and serialized then
+    /// this method can be useful to get the serialized version without
+    /// compiling twice.
     #[cfg(compiler)]
     #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn serialize(&self) -> Result<Vec<u8>> {
