@@ -601,6 +601,12 @@ impl<'a> SerializedModule<'a> {
 
             // This doesn't affect compilation, it's just a runtime setting.
             dynamic_memory_growth_reserve: _,
+
+            // This does technically affect compilation but modules with/without
+            // trap information can be loaded into engines with the opposite
+            // setting just fine (it's just a section in the compiled file and
+            // whether it's present or not)
+            generate_address_map: _,
         } = self.metadata.tunables;
 
         Self::check_int(
@@ -805,26 +811,6 @@ mod test {
         Ok(())
     }
 
-    #[cfg(feature = "lightbeam")]
-    #[test]
-    fn test_compilation_strategy_mismatch() -> Result<()> {
-        let engine = Engine::default();
-        let module = Module::new(&engine, "(module)")?;
-
-        let mut serialized = SerializedModule::new(&module);
-        serialized.metadata.strategy = CompilationStrategy::Lightbeam;
-
-        match serialized.into_module(&engine) {
-            Ok(_) => unreachable!(),
-            Err(e) => assert_eq!(
-                e.to_string(),
-                "Module was compiled with strategy 'Cranelift'",
-            ),
-        }
-
-        Ok(())
-    }
-
     #[test]
     fn test_tunables_int_mismatch() -> Result<()> {
         let engine = Engine::default();
