@@ -119,7 +119,7 @@ impl fmt::Display for TrapCode {
             IntegerOverflow => "integer overflow",
             IntegerDivisionByZero => "integer divide by zero",
             BadConversionToInteger => "invalid conversion to integer",
-            UnreachableCodeReached => "unreachable",
+            UnreachableCodeReached => "wasm `unreachable` instruction executed",
             Interrupt => "interrupt",
         };
         write!(f, "{}", desc)
@@ -321,7 +321,11 @@ impl fmt::Display for Trap {
         writeln!(f, "\nwasm backtrace:")?;
         for (i, frame) in self.trace().iter().enumerate() {
             let name = frame.module_name().unwrap_or("<unknown>");
-            write!(f, "  {:>3}: {:#6x} - ", i, frame.module_offset())?;
+            write!(f, "  {:>3}: ", i)?;
+
+            if let Some(offset) = frame.module_offset() {
+                write!(f, "{:#6x} - ", offset)?;
+            }
 
             let demangle =
                 |f: &mut fmt::Formatter<'_>, name: &str| match rustc_demangle::try_demangle(name) {
@@ -385,7 +389,10 @@ impl std::error::Error for Trap {
 
 impl From<anyhow::Error> for Trap {
     fn from(e: anyhow::Error) -> Trap {
-        Box::<dyn std::error::Error + Send + Sync>::from(e).into()
+        match e.downcast::<Trap>() {
+            Ok(trap) => trap,
+            Err(e) => Box::<dyn std::error::Error + Send + Sync>::from(e).into(),
+        }
     }
 }
 
